# Copyright (c) 2021 Hongji Wang (jijijiang77@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from pprint import pformat

import fire
import tableprint as tp
import torch
import torch.distributed as dist
import yaml
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

import dguard.speaker.utils.schedulers as schedulers
from dguard.speaker.dataset.dataset_deprecated import FeatList_LableDict_Dataset
from dguard.speaker.models.projections import get_projection
from dguard.speaker.models.speaker_model import get_speaker_model
from dguard.speaker.utils.checkpoint import load_checkpoint, save_checkpoint
from dguard.speaker.utils.executor_deprecated import run_epoch
from dguard.speaker.utils.file_utils import read_scp
from dguard.speaker.utils.utils import (
    get_logger,
    parse_config_or_kwargs,
    set_seed,
    spk2id,
)


def train(config="conf/config.yaml", **kwargs):
    """Trains a model on the given features and spk labels.

    :config: A training configuration. Note that all parameters in the
             config can also be manually adjusted with --ARG VALUE
    :returns: None
    """

    configs = parse_config_or_kwargs(config, **kwargs)
    checkpoint = configs.get("checkpoint", None)
    # dist configs
    rank = int(os.environ["LOCAL_RANK"])
    world_size = int(os.environ["WORLD_SIZE"])
    gpu = int(configs["gpus"][rank])
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu)
    dist.init_process_group(backend="nccl")

    model_dir = os.path.join(configs["exp_dir"], "models")
    if rank == 0:
        try:
            os.makedirs(model_dir)
        except IOError:
            print(model_dir + " already exists !!!")
            if checkpoint is None:
                exit(1)
    dist.barrier()  # let the rank 0 mkdir first

    logger = get_logger(configs["exp_dir"], "train.log")
    if world_size > 1:
        logger.info("training on multiple gpus, this gpu {}".format(gpu))

    if rank == 0:
        logger.info("exp_dir is: {}".format(configs["exp_dir"]))
        logger.info("<== Passed Arguments ==>")
        # Print arguments into logs
        for line in pformat(configs).split("\n"):
            logger.info(line)

    # seed
    set_seed(configs["seed"] + rank)

    # wav/feat
    train_scp = configs["dataset_args"]["train_scp"]
    train_label = configs["dataset_args"]["train_label"]
    train_data_list = read_scp(train_scp)
    if rank == 0:
        logger.info("<== Feature ==>")
        logger.info("train wav/feat num: {}".format(len(train_data_list)))

    # spk label
    train_utt_spk_list = read_scp(train_label)
    spk2id_dict = spk2id(train_utt_spk_list)
    train_utt2spkid_dict = {
        utt_spk[0]: spk2id_dict[utt_spk[1]] for utt_spk in train_utt_spk_list
    }
    if rank == 0:
        logger.info("<== Labels ==>")
        logger.info(
            "train label num: {}, spk num: {}".format(
                len(train_utt2spkid_dict), len(spk2id_dict)
            )
        )

    # dataset and dataloader
    configs["feature_args"]["feat_dim"] = configs["model_args"]["feat_dim"]
    train_dataset = FeatList_LableDict_Dataset(
        train_data_list,
        train_utt2spkid_dict,
        **configs["feature_args"],
        **configs["dataset_args"]
    )
    train_sampler = DistributedSampler(train_dataset, shuffle=True)
    train_dataloader = DataLoader(
        train_dataset, sampler=train_sampler, **configs["dataloader_args"]
    )
    if rank == 0:
        logger.info("<== Dataloaders ==>")
        logger.info("train dataloaders created")

    # model
    logger.info("<== Model ==>")
    model = get_speaker_model(configs["model"])(**configs["model_args"])
    if configs["model_init"] is not None:
        logger.info("Load initial model from {}".format(configs["model_init"]))
        load_checkpoint(model, configs["model_init"])
    else:
        logger.info("Train model from scratch...")
    # projection layer
    configs["projection_args"]["embed_dim"] = configs["model_args"]["embed_dim"]
    configs["projection_args"]["num_class"] = len(spk2id_dict)
    if configs["feature_args"]["raw_wav"] and configs["dataset_args"]["speed_perturb"]:
        # diff speed is regarded as diff spk
        configs["projection_args"]["num_class"] *= 3
    configs["projection_args"]["do_lm"] = configs.get("do_lm", False)
    projection = get_projection(configs["projection_args"])
    model.add_module("projection", projection)
    if rank == 0:
        # print model
        for line in pformat(model).split("\n"):
            logger.info(line)
        # !!!IMPORTANT!!!
        # Try to export the model by script, if fails, we should refine
        # the code to satisfy the script export requirements
        script_model = torch.jit.script(model)
        script_model.save(os.path.join(model_dir, "init.zip"))

    # If specify checkpoint, load some info from checkpoint.
    if checkpoint is not None:
        load_checkpoint(model, checkpoint)
        start_epoch = int(re.findall(r"(?<=model_)\d*(?=.pt)", checkpoint)[0]) + 1
        logger.info("checkpoint: {}".format(checkpoint))
    else:
        start_epoch = 1
    logger.info("start_epoch: {}".format(start_epoch))

    # ddp_model
    model.cuda()
    ddp_model = torch.nn.parallel.DistributedDataParallel(model)
    device = torch.device("cuda")

    criterion = getattr(torch.nn, configs["loss"])(**configs["loss_args"])
    if rank == 0:
        logger.info("<== Loss ==>")
        logger.info("loss criterion is: " + configs["loss"])

    configs["optimizer_args"]["lr"] = configs["scheduler_args"]["initial_lr"]
    optimizer = getattr(torch.optim, configs["optimizer"])(
        ddp_model.parameters(), **configs["optimizer_args"]
    )
    if rank == 0:
        logger.info("<== Optimizer ==>")
        logger.info("optimizer is: " + configs["optimizer"])

    # scheduler
    configs["scheduler_args"]["num_epochs"] = configs["num_epochs"]
    configs["scheduler_args"]["epoch_iter"] = len(train_dataloader)
    scheduler = getattr(schedulers, configs["scheduler"])(
        optimizer, **configs["scheduler_args"]
    )
    if rank == 0:
        logger.info("<== Scheduler ==>")
        logger.info("scheduler is: " + configs["scheduler"])

    # margin scheduler
    configs["margin_update"]["epoch_iter"] = len(train_dataloader)
    margin_scheduler = getattr(schedulers, configs["margin_scheduler"])(
        model=model, **configs["margin_update"]
    )
    if rank == 0:
        logger.info("<== MarginScheduler ==>")

    # save config.yaml
    if rank == 0:
        saved_config_path = os.path.join(configs["exp_dir"], "config.yaml")
        with open(saved_config_path, "w") as fout:
            data = yaml.dump(configs)
            fout.write(data)

    # training
    dist.barrier()  # synchronize here
    if rank == 0:
        logger.info("<========== Training process ==========>")
        header = ["Epoch", "Batch", "Lr", "Margin", "Loss", "Acc"]
        for line in tp.header(header, width=10, style="grid").split("\n"):
            logger.info(line)
    dist.barrier()  # synchronize here

    for epoch in range(start_epoch, configs["num_epochs"] + 1):
        train_sampler.set_epoch(epoch)

        run_epoch(
            train_dataloader,
            ddp_model,
            criterion,
            optimizer,
            scheduler,
            margin_scheduler,
            epoch,
            logger,
            log_batch_interval=configs["log_batch_interval"],
            device=device,
        )

        if rank == 0:
            if (
                epoch % configs["save_epoch_interval"] == 0
                or epoch >= configs["num_epochs"] - configs["num_avg"]
            ):
                save_checkpoint(
                    model, os.path.join(model_dir, "model_{}.pt".format(epoch))
                )

    if rank == 0:
        os.symlink(
            "model_{}.pt".format(configs["num_epochs"]),
            os.path.join(model_dir, "final_model.pt"),
        )
        logger.info(tp.bottom(len(header), width=10, style="grid"))


if __name__ == "__main__":
    fire.Fire(train)
